@@ -20,6 +20,8 @@
 #include <stdint.h>
 #include <assert.h>
 
+#include "utils.h"
+
 #ifdef assert
 #define RansAssert assert
 #else
@@ -128,7 +130,7 @@ static inline void RansDecInit(RansState* r, uint8_t** pptr)
     x  = ptr[0] << 0;
     x |= ptr[1] << 8;
     x |= ptr[2] << 16;
-    x |= ptr[3] << 24;
+    x |= ((uint32_t)ptr[3]) << 24;
     ptr += 4;
 
     *pptr = ptr;
@@ -286,14 +288,20 @@ static inline void RansEncPutSymbol(RansState* r, uint8_t** pptr, RansEncSymbol
     uint32_t x = *r;
     uint32_t x_max = sym->x_max;
 
-    if (x >= x_max) {
-	uint8_t* ptr = *pptr;
-	do {
-	    *--ptr = (uint8_t) (x & 0xff);
-	    x >>= 8;
-	} while (x >= x_max);
-	*pptr = ptr;
+    // This is better for 40-qual illumina (3.7% quicker overall CRAM).
+    // The old method was better for low complexity data such as NovaSeq
+    // quals (2.6% quicker overall CRAM).
+    int o = x >= x_max;
+    uint8_t* ptr = *pptr;
+    ptr[-1] = x & 0xff;
+    ptr -= o;
+    x >>= o*8;
+
+    if (unlikely(x >= x_max)) {
+        *--ptr = (uint8_t) (x & 0xff);
+        x >>= 8;
     }
+    *pptr = ptr;
 
     //uint32_t q = (uint32_t) (((uint64_t)x * sym->rcp_freq) >> sym->rcp_shift);
     //*r = q * sym->cmpl_freq + x + sym->bias;
@@ -309,6 +317,97 @@ static inline void RansEncPutSymbol(RansState* r, uint8_t** pptr, RansEncSymbol
     *r = q * sym->cmpl_freq + x + sym->bias;
 }
 
+// A 4-way version of RansEncPutSymbol, renormalising 4 states
+// simulatenously with their results written to the same ptr buffer.
+// (This is perhaps a failing as it makes optmisation tricky.)
+static inline void RansEncPutSymbol4(RansState *r0,
+                                     RansState *r1,
+                                     RansState *r2,
+                                     RansState *r3,
+                                     uint8_t** pptr,
+                                     RansEncSymbol const *sym0,
+                                     RansEncSymbol const *sym1,
+                                     RansEncSymbol const *sym2,
+                                     RansEncSymbol const *sym3)
+{
+    RansAssert(sym0->x_max != 0); // can't encode symbol with freq=0
+    RansAssert(sym1->x_max != 0); // can't encode symbol with freq=0
+    RansAssert(sym2->x_max != 0); // can't encode symbol with freq=0
+    RansAssert(sym3->x_max != 0); // can't encode symbol with freq=0
+
+    // renormalize
+    uint32_t x0, x1, x2, x3;
+    uint8_t* ptr = *pptr;
+
+    int o;
+    uint32_t m[4] = {
+        sym0->x_max,
+        sym1->x_max,
+        sym2->x_max,
+        sym3->x_max
+    };
+
+    x0 = *r0;
+    o = x0 >= m[0];
+    ptr[-1] = x0;
+    ptr -= o;
+    x0 >>= o*8;
+    if (x0 >= m[0]) {
+        *--ptr = x0;
+        x0 >>= 8;
+    }
+
+    x1 = *r1;
+    o = x1 >= m[1];
+    ptr[-1] = x1;
+    ptr -= o;
+    x1 >>= o*8;
+    if (x1 >= m[1]) {
+        *--ptr = x1;
+        x1 >>= 8;
+    }
+
+    x2 = *r2;
+    o = x2 >= m[2];
+    ptr[-1] = x2;
+    ptr -= o;
+    x2 >>= o*8;
+    if (x2 >= m[2]) {
+        *--ptr = x2;
+        x2 >>= 8;
+    }
+
+    x3 = *r3;
+    o = x3 >= m[3];
+    ptr[-1] = x3;
+    ptr -= o;
+    x3 >>= o*8;
+    if (x3 >= m[3]) {
+        *--ptr = x3;
+        x3 >>= 8;
+    }
+
+    *pptr = ptr;
+
+    // x = C(s,x)
+    uint32_t qa, qb;
+    qa = (uint32_t) (((uint64_t)x0 * sym0->rcp_freq) >> sym0->rcp_shift);
+    uint32_t X0 = qa * sym0->cmpl_freq;
+    qb = (uint32_t) (((uint64_t)x1 * sym1->rcp_freq) >> sym1->rcp_shift);
+    uint32_t X1 = qb * sym1->cmpl_freq;
+
+    *r0 = X0 + x0 + sym0->bias;
+    *r1 = X1 + x1 + sym1->bias;
+
+    qa = (uint32_t) (((uint64_t)x2 * sym2->rcp_freq) >> sym2->rcp_shift);
+    uint32_t X2 = qa * sym2->cmpl_freq;
+    qb = (uint32_t) (((uint64_t)x3 * sym3->rcp_freq) >> sym3->rcp_shift);
+    uint32_t X3 = qb * sym3->cmpl_freq;
+
+    *r2 = X2 + x2 + sym2->bias;
+    *r3 = X3 + x3 + sym3->bias;
+}
+
 // Equivalent to RansDecAdvance that takes a symbol.
 static inline void RansDecAdvanceSymbol(RansState* r, uint8_t** pptr, RansDecSymbol const* sym, uint32_t scale_bits)
 {
@@ -345,8 +444,8 @@ static inline void RansDecRenorm(RansState* r, uint8_t** pptr) {
     uint8_t  *ptr = *pptr;
 
     __asm__ ("movzbl (%0), %%eax\n\t"
-	     "mov    %1, %%edx\n\t"
-	     "shl    $0x8,%%edx\n\t"
+             "mov    %1, %%edx\n\t"
+             "shl    $0x8,%%edx\n\t"
              "or     %%eax,%%edx\n\t"
              "cmp    $0x800000,%1\n\t"
              "cmovb  %%edx,%1\n\t"
@@ -449,7 +548,7 @@ static inline void RansDecRenormSafe(RansState* r, uint8_t** pptr, uint8_t *ptr_
     if (x >= RANS_BYTE_L || ptr >= ptr_end) return;
     x = (x << 8) | *ptr++;
     if (x < RANS_BYTE_L && ptr < ptr_end)
-	x = (x << 8) | *ptr++;
+        x = (x << 8) | *ptr++;
     *pptr = ptr;
     *r = x;
 }
